from sonusai.mixture.truth_functions.data import Data
from sonusai.mixture.types import Truth


def target_f(data: Data) -> Truth:
    import numpy as np

    from sonusai import SonusAIError

    if data.config.num_classes != 2 * data.target_fft.bins:
        raise SonusAIError(f'Invalid num_classes for target_f truth: {data.config.num_classes}')

    for offset in data.offsets:
        target_freq = np.complex64(data.target_fft.execute(data.target_audio[offset:offset + data.frame_size]))

        indices = slice(offset, offset + data.frame_size)
        for index in data.zero_based_indices:
            start = index
            stop = start + data.target_fft.bins
            data.truth[indices, start:stop] = np.real(target_freq)

            start = stop
            stop = start + data.target_fft.bins
            data.truth[indices, start:stop] = np.imag(target_freq)

    return data.truth


def target_mixture_f(data: Data) -> Truth:
    import numpy as np

    from sonusai import SonusAIError

    if data.config.num_classes != 2 * data.target_fft.bins + 2 * data.noise_fft.bins:
        raise SonusAIError(f'Invalid num_classes for target_mixture_f truth: {data.config.num_classes}')

    for offset in data.offsets:
        target_freq = np.complex64(data.target_fft.execute(data.target_audio[offset:offset + data.frame_size]))
        noise_freq = np.complex64(data.noise_fft.execute(data.noise_audio[offset:offset + data.frame_size]))
        mixture_freq = target_freq + noise_freq

        indices = slice(offset, offset + data.frame_size)
        for index in data.zero_based_indices:
            start = index
            stop = start + data.target_fft.bins
            data.truth[indices, start:stop] = np.real(target_freq)

            start = stop
            stop = start + data.target_fft.bins
            data.truth[indices, start:stop] = np.imag(target_freq)

            start = stop
            stop = start + data.noise_fft.bins
            data.truth[indices, start:stop] = np.real(mixture_freq)

            start = stop
            stop = start + data.noise_fft.bins
            data.truth[indices, start:stop] = np.imag(mixture_freq)

    return data.truth


def target_swin_f(data: Data) -> Truth:
    # import numpy as np

    from sonusai import SonusAIError

    # if data.config.num_classes != 2 * data.target_fft.bins:
    #     raise SonusAIError(f'Invalid num_classes for target_swin_f truth: {data.config.num_classes}')
    #
    # for offset in data.offsets:
    #     target_freq = np.complex64(
    #         data.target_fft.execute(np.multiply(data.target_audio[offset:offset + data.frame_size], data.swin)))
    #
    #     indices = slice(offset, offset + data.frame_size)
    #     for index in data.zero_based_indices:
    #         start = index
    #         stop = start + data.target_fft.bins
    #         data.truth[indices, start:stop] = np.real(target_freq)
    #
    #         start = stop
    #         stop = start + data.target_fft.bins
    #         data.truth[indices, start:stop] = np.imag(target_freq)
    #
    # return data.truth

    raise SonusAIError('Truth function target_swin_f is not supported yet')



from sonusai.utils.asr_functions.aixplain_whisper import aixplain_whisper
from sonusai.utils.asr_functions.deepgram import deepgram
from sonusai.utils.asr_functions.google import google
from sonusai.utils.asr_functions.whisper import whisper



# SonusAI truth functions
from sonusai.mixture.truth_functions.crm import crm
from sonusai.mixture.truth_functions.crm import crmp
from sonusai.mixture.truth_functions.energy import energy_f
from sonusai.mixture.truth_functions.energy import energy_t
from sonusai.mixture.truth_functions.energy import mapped_snr_f
from sonusai.mixture.truth_functions.energy import snr_f
from sonusai.mixture.truth_functions.file import file
from sonusai.mixture.truth_functions.phoneme import phoneme
from sonusai.mixture.truth_functions.sed import sed
from sonusai.mixture.truth_functions.target import target_f
from sonusai.mixture.truth_functions.target import target_mixture_f
from sonusai.mixture.truth_functions.target import target_swin_f



from typing import Union

import numpy as np
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.data import Sampler

from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase


class MixtureDatabaseDataset(Dataset):
    """Generates a PyTorch dataset from a SonusAI mixture database
    """

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: GeneralizedIDs,
                 cut_len: int,
                 flatten: bool,
                 add1ch: bool,
                 random_cut: bool = True):
        """Initialization
        """
        self.mixdb = mixdb
        self.mixids = self.mixdb.mixids_to_list(mixids)
        self.cut_len = cut_len
        self.flatten = flatten
        self.add1ch = add1ch
        self.random_cut = random_cut

    def __len__(self):
        return len(self.mixids)

    def __getitem__(self, idx: int) -> (np.ndarray, np.ndarray, int):
        """Get data from one mixture
        """
        import random

        from sonusai.utils import reshape_inputs

        feature, truth = self.mixdb.mixture_ft(self.mixids[idx])
        feature, truth = reshape_inputs(feature=feature,
                                        truth=truth,
                                        batch_size=1,
                                        timesteps=0,
                                        flatten=self.flatten,
                                        add1ch=self.add1ch)

        length = feature.shape[0]

        if self.cut_len > 0:
            if length < self.cut_len:
                feature_final = []
                truth_final = []
                for _ in range(self.cut_len // length):
                    feature_final.append(feature)
                    truth_final.append(truth)
                feature_final.append(feature[: self.cut_len % length])
                truth_final.append(truth[: self.cut_len % length])
                feature = np.vstack([feature_final[i] for i in range(len(feature_final))])
                truth = np.vstack([truth_final[i] for i in range(len(truth_final))])
            else:
                if self.random_cut:
                    start = random.randint(0, length - self.cut_len)
                else:
                    start = 0
                feature = feature[start:start + self.cut_len]
                truth = truth[start:start + self.cut_len]

        return truth, feature, length


def TorchFromMixtureDatabase(mixdb: MixtureDatabase,
                             mixids: GeneralizedIDs,
                             batch_size: int,
                             flatten: bool,
                             add1ch: bool,
                             num_workers: int = 0,
                             cut_len: int = 0,
                             drop_last: bool = False,
                             shuffle: bool = False,
                             random_cut: bool = True,
                             sampler: Union[Sampler, None] = None) -> DataLoader:
    """Generates a PyTorch dataloader from a SonusAI mixture database
    """
    dataset = MixtureDatabaseDataset(mixdb=mixdb,
                                     mixids=mixids,
                                     cut_len=cut_len,
                                     flatten=flatten,
                                     add1ch=add1ch,
                                     random_cut=random_cut)

    if sampler is not None:
        sampler = sampler(dataset)

    result = DataLoader(dataset=dataset,
                        batch_size=batch_size,
                        pin_memory=True,
                        shuffle=shuffle,
                        sampler=sampler,
                        drop_last=drop_last,
                        num_workers=num_workers)

    result.cut_len = cut_len

    return result



"""sonusai calc_metric_spenh_targetf

usage: calc_metric_spenh_targetf [-hvtpws] [-i MIXID] (-d PLOC) [-e WER] INPUT

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate. [default: *].
    -d PLOC, --ploc PLOC        Location of SonusAI predict data.
    -e WER, --wer-method WER    Word-Error-Rate method: deepgram, google, or whisper [default: none]
    -t, --truth-est-mode        Calculate extraction using truth and include metrics.
    -p, --plot                  Enable PDF plots file generation per mixture.
    -w, --wav                   Enable WAV file generation per mixture.
    -s, --summary               Enable summary files generation.

Calculate speech enhancement target_f metrics for prediction data in PLOC and SonusAI mixture database in INPUT.

Inputs:
    PLOC    SonusAI prediction data directory.
    INPUT   SonusAI mixture database directory.

Outputs the following to PLOC:
    <id>_metric_spenh_targetf.txt

    If --plot:
        <id>_metric_spenh_targetf.pdf

    If --wav:
        <id>_target.wav
        <id>_target_est.wav
        <id>_noise.wav
        <id>_noise_est.wav
        <id>_mixture.wav

        If --truth-est-mode:
            <id>_target_truth_est.wav
            <id>_noise_truth_est.wav

    If --summary:
        metric_spenh_targetf_summary.txt
        metric_spenh_targetf_summary.csv
        metric_spenh_targetf_list.csv
        metric_spenh_targetf_estats_list.csv

        If --truth-est-mode:
            metric_spenh_targetf_truth_list.csv
            metric_spenh_targetf_estats_truth_list.csv

"""
from dataclasses import dataclass
from typing import Tuple
from typing import Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sonusai import logger
from sonusai.mixture import AudioF
from sonusai.mixture import AudioT
from sonusai.mixture import Feature
from sonusai.mixture import Location
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Predict

matplotlib.use('SVG')


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    predict_location: Location = None
    truth_est_mode: bool = None
    enable_plot: bool = None
    enable_wav: bool = None
    wer_method: str = None


MP_GLOBAL = MPGlobal()


def mean_square_error(hypothesis: np.ndarray,
                      reference: np.ndarray,
                      squared: bool = False) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate root mean square error or mean square error

    :param hypothesis: [frames, bins]
    :param reference: [frames, bins]
    :param squared: calculate mean square rather than root mean square
    :return: mean, mean per bin, mean per frame
    """
    sq_err = np.square(reference - hypothesis)

    # mean over frames for value per bin
    err_b = np.mean(sq_err, axis=0)
    # mean over bins for value per frame
    err_f = np.mean(sq_err, axis=1)
    # mean over all
    err = np.mean(sq_err)

    if not squared:
        err_b = np.sqrt(err_b)
        err_f = np.sqrt(err_f)
        err = np.sqrt(err)

    return err, err_b, err_f


def mean_abs_percentage_error(hypothesis: np.ndarray, reference: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate mean abs percentage error

    If inputs are complex, calculates average: mape(real)/2 + mape(imag)/2

    :param hypothesis: [frames, bins]
    :param reference: [frames, bins]
    :return: mean, mean per bin, mean per frame
    """
    if not np.iscomplexobj(reference) and not np.iscomplexobj(hypothesis):
        abs_err = 100 * np.abs((reference - hypothesis) / (reference + np.finfo(np.float32).eps))
    else:
        reference_r = np.real(reference)
        reference_i = np.imag(reference)
        hypothesis_r = np.real(hypothesis)
        hypothesis_i = np.imag(hypothesis)
        abs_err_r = 100 * np.abs((reference_r - hypothesis_r) / (reference_r + np.finfo(np.float32).eps))
        abs_err_i = 100 * np.abs((reference_i - hypothesis_i) / (reference_i + np.finfo(np.float32).eps))
        abs_err = (abs_err_r / 2) + (abs_err_i / 2)

    # mean over frames for value per bin
    err_b = np.around(np.mean(abs_err, axis=0), 3)
    # mean over bins for value per frame
    err_f = np.around(np.mean(abs_err, axis=1), 3)
    # mean over all
    err = np.around(np.mean(abs_err), 3)

    return err, err_b, err_f


def log_error(reference: np.ndarray, hypothesis: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate log error

    :param reference: complex or real [frames, bins]
    :param hypothesis: complex or real [frames, bins]
    :return: mean, mean per bin, mean per frame
    """
    reference_sq = np.real(reference * np.conjugate(reference))
    hypothesis_sq = np.real(hypothesis * np.conjugate(hypothesis))
    log_err = abs(10 * np.log10((reference_sq + np.finfo(np.float32).eps) / (hypothesis_sq + np.finfo(np.float32).eps)))
    # log_err = abs(10 * np.log10(reference_sq / (hypothesis_sq + np.finfo(np.float32).eps) + np.finfo(np.float32).eps))

    # mean over frames for value per bin
    err_b = np.around(np.mean(log_err, axis=0), 3)
    # mean over bins for value per frame
    err_f = np.around(np.mean(log_err, axis=1), 3)
    # mean over all
    err = np.around(np.mean(log_err), 3)

    return err, err_b, err_f


def plot_mixpred(mixture: AudioT,
                 mixture_f: AudioF,
                 target: AudioT = None,
                 feature: Feature = None,
                 predict: Predict = None,
                 tp_title: str = '') -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE

    num_plots = 2
    if feature is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, mixture, label='Mixture', color='mistyrose')
    ax[0].set_ylabel('magnitude', color='tab:blue')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    # ax[p].set_ylim([-1.025, 1.025])
    if target is not None:  # Plot target time-domain waveform on top of mixture
        ax[0].plot(x_axis, target, label='Target', color='tab:blue')
        # ax[0].tick_params(axis='y', labelcolor=color)
    ax[p].set_title('Waveform')

    # Plot the mixture spectrogram
    p += 1
    ax[p].imshow(np.transpose(mixture_f), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Mixture')

    if feature is not None:
        p += 1
        ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Feature')

    if predict is not None:
        p += 1
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict ' + tp_title)

    return fig


def plot_pdb_predtruth(predict: np.ndarray,
                       truth_f: Union[np.ndarray, None] = None,
                       metric: Union[np.ndarray, None] = None,
                       tp_title: str = '') -> plt.figure:
    """Plot predict and optionally truth and a metric in power db, e.g. applies 10*log10(predict)"""
    num_plots = 2
    if truth_f is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    tmp = 10 * np.log10(predict.transpose() + np.finfo(np.float32).eps)
    ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')

    if truth_f is not None:
        p += 1
        tmp = 10 * np.log10(truth_f.transpose() + np.finfo(np.float32).eps)
        ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict avg, and optionally truth avg and metric lines
    pred_avg = 10 * np.log10(np.mean(predict, axis=-1) + np.finfo(np.float32).eps)
    p += 1
    x_axis = np.arange(len(pred_avg), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, pred_avg, color='black', linestyle='dashed', label='Predict mean over freq.')
    ax[p].set_ylabel('mean db', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_f is not None:
        truth_avg = 10 * np.log10(np.mean(truth_f, axis=-1) + np.finfo(np.float32).eps)
        ax[p].plot(x_axis, truth_avg, color='green', linestyle='dashed', label='Truth mean over freq.')

    if metric is not None:  # instantiate 2nd y-axis that shares the same x-axis
        ax2 = ax[p].twinx()
        color2 = 'red'
        ax2.plot(x_axis, metric, color=color2, label='sig distortion (mse db)')
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([0, np.max(metric)])
        ax2.set_ylabel('spectral distortion (mse db)', color=color2)
        ax2.tick_params(axis='y', labelcolor=color2)
        ax[p].set_title('SNR and SNR mse (mean over freq. db)')
    else:
        ax[p].set_title('SNR (mean over freq. db)')
        # ax[0].tick_params(axis='y', labelcolor=color)
    return fig


def plot_epredtruth(predict: np.ndarray,
                    predict_wav: np.ndarray,
                    truth_f: Union[np.ndarray, None] = None,
                    truth_wav: Union[np.ndarray, None] = None,
                    metric: Union[np.ndarray, None] = None,
                    tp_title: str = '') -> plt.figure:
    """Plot predict spectrogram and waveform and optionally truth and a metric)"""
    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if metric is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    ax[p].imshow(predict.transpose(), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')

    if truth_f is not None:
        p += 1
        ax[p].imshow(truth_f.transpose(), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict wav, and optionally truth avg and metric lines
    p += 1
    x_axis = np.arange(len(predict_wav), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, predict_wav, color='black', linestyle='dashed', label='Speech Estimate')
    ax[p].set_ylabel('Amplitude', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_wav is not None:
        ntrim = len(truth_wav) - len(predict_wav)
        if ntrim > 0:
            truth_wav = truth_wav[0:-ntrim]
        ax[p].plot(x_axis, truth_wav, color='green', linestyle='dashed', label='True Target')

    # Plot the metric lines
    if metric is not None:
        p += 1
        x_axis = np.arange(len(metric), dtype=np.float32)  # / SAMPLE_RATE
        ax[p].plot(x_axis, metric, color='red', label='Target LogErr')
        ax[p].set_ylabel('log error db', color='red')
        ax[p].set_xlim(x_axis[0], x_axis[-1])
        ax[p].set_ylim([-0.01, np.max(metric) + .01])

    return fig


def _process_mixture(mixid: int) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    from os.path import basename
    from os.path import join
    from os.path import splitext

    import h5py

    from sonusai import SonusAIError
    from sonusai.metrics import calc_pcm
    from sonusai.metrics import calc_pesq
    from sonusai.metrics import calc_sa_sdr
    from sonusai.metrics import calc_wer
    from sonusai.utils import calc_asr
    from sonusai.utils import float_to_int16
    from sonusai.utils import unstack_complex
    from sonusai.utils import write_wav

    mixdb = MP_GLOBAL.mixdb
    predict_location = MP_GLOBAL.predict_location
    truth_est_mode = MP_GLOBAL.truth_est_mode
    enable_plot = MP_GLOBAL.enable_plot
    enable_wav = MP_GLOBAL.enable_wav
    wer_method = MP_GLOBAL.wer_method

    # 1) Collect true target, noise, mixture data
    target = mixdb.mixture_target(mixid)
    target_f = mixdb.mixture_target_f(mixid, target=target)

    noise = mixdb.mixture_noise(mixid)
    noise_f = mixdb.mixture_noise_f(mixid, noise=noise)

    mixture = mixdb.mixture_mixture(mixid, target=target, noise=noise)
    mixture_f = mixdb.mixture_mixture_f(mixid, mixture=mixture)

    feature, truth_f = mixdb.mixture_ft(mixid, mixture=mixture)

    # need to use inv-tf to match size & tf properties
    targetfi = mixdb.inverse_transform(target_f)
    noisefi = mixdb.inverse_transform(noise_f)

    # Test code for tf-based aaware inv transform
    # mcf = np.concatenate([np.expand_dims(target_f,0), np.expand_dims(noise_f,0)],axis=0)
    # ttf = tf_istft_olsa_hanns(mcf, N=mixdb.it_config.N,
    #                           R=mixdb.it_config.R,
    #                           BinStart=mixdb.fg.bin_start,
    #                           BinEnd=mixdb.fg.bin_end)
    # write_wav('tf_target.wav', audio=float_to_int16(ttf[0,:]))
    # write_wav('tf_noise.wav', audio=float_to_int16(ttf[1, :]))
    # to = tf_istft_olsa_hanns(target_f, N=mixdb.it_config.N,
    #                           R=mixdb.it_config.R,
    #                           BinStart=mixdb.fg.bin_start,
    #                           BinEnd=mixdb.fg.bin_end)
    # write_wav('tf_target_only.wav', audio=float_to_int16(to))

    # 2)  Read predict data
    output_name = join(predict_location, mixdb.mixtures[mixid].name)
    base_name = splitext(output_name)[0]
    try:
        with h5py.File(output_name, 'r') as f:
            predict = np.array(f['predict'])
    except Exception as e:
        raise SonusAIError(f'Error reading {output_name}: {e}')

    # 3) Extraction - for target_f truth, simply unstack as predict is estimating the target
    predict_complex = unstack_complex(predict)
    truth_f_complex = unstack_complex(truth_f)
    noise_est_complex = mixture_f - predict_complex
    target_est_wav = mixdb.inverse_transform(predict_complex)
    noise_est_wav = mixdb.inverse_transform(noise_est_complex)

    target_truth_est_complex = None
    target_truth_est_audio = None

    noise_truth_est_complex = None
    noise_truth_est_audio = None

    if truth_est_mode:
        # estimates using truth instead of prediction
        target_truth_est_complex = truth_f_complex
        target_truth_est_audio = mixdb.inverse_transform(target_truth_est_complex)

        noise_truth_est_complex = mixture_f - target_truth_est_complex
        noise_truth_est_audio = mixdb.inverse_transform(noise_truth_est_complex)

    # 4) Metrics
    # Mean absolute percentage error (real and imag)
    # cmape_tg, cmape_tg_bin, cmape_tg_frame = mean_abs_percentage_error(hypothesis=predict_complex,
    #                                                                    reference=truth_f_complex)

    # Target/Speech logerr - PSD estimation accuracy symmetric mean log-spectral distortion
    lerr_tg, lerr_tg_bin, lerr_tg_frame = log_error(truth_f_complex, predict_complex)
    # Noise logerr - PSD estimation accuracy
    lerr_n, lerr_n_bin, lerr_n_frame = log_error(noise_f, noise_est_complex)
    # PCM loss metric
    ytrue_f = np.concatenate((truth_f_complex[:, np.newaxis, :], noise_f[:, np.newaxis, :]), axis=1)
    ypred_f = np.concatenate((predict_complex[:, np.newaxis, :], noise_est_complex[:, np.newaxis, :]), axis=1)
    pcm, pcm_bin, pcm_frame = calc_pcm(hypothesis=ypred_f, reference=ytrue_f, with_log=True)

    # Noise td logerr
    # lerr_nt, lerr_nt_bin, lerr_nt_frame = log_error(noisefi, noise_truth_est_audio)

    # SA-SDR (time-domain source-aggragated SDR)
    ytrue = np.concatenate((targetfi[:, np.newaxis], noisefi[:, np.newaxis]), axis=1)
    ypred = np.concatenate((target_est_wav[:, np.newaxis], noise_est_wav[:, np.newaxis]), axis=1)
    # note: w/o scale is more pessimistic number
    sa_sdr, _ = calc_sa_sdr(hypothesis=ypred, reference=ytrue)

    # Speech intelligibility measure - PESQ
    if int(mixdb.mixtures[mixid].snr) > -99:
        pesq_speech = calc_pesq(hypothesis=target_est_wav, reference=target)
        pesq_mixture = calc_pesq(hypothesis=mixture, reference=target)
        # pesq improvement
        pesq_impr = pesq_speech - pesq_mixture
        # pesq improvement %
        pesq_impr_pc = pesq_impr / (pesq_mixture + np.finfo(np.float32).eps) * 100
    else:
        pesq_speech = 0
        pesq_mixture = 0
        pesq_impr_pc = 0

    # Calc WER
    if wer_method == 'none':
        wer_mx = float('nan')
        wer_tge = float('nan')
        wer_pi = float('nan')
    else:
        # i.e., wer_method ='google'
        asr_tt = calc_asr(target, engine=wer_method)  # target truth
        asr_mx = calc_asr(mixture, engine=wer_method)
        asr_tge = calc_asr(target_est_wav, engine=wer_method)

        tmp = calc_wer(asr_mx.text, asr_tt.text)  # mixture wer
        wer_mx = tmp.wer * 100
        tmp = calc_wer(asr_tge.text, asr_tt.text)  # target estimate wer
        wer_tge = tmp.wer * 100
        wer_pi = (wer_mx - wer_tge) / (wer_mx + 0.1) * 100

    # 5) Save per mixture metric results
    # Single row in table of scalar metrics per mixture
    mtable1_col = ['MXSNR', 'MXPESQ', 'MXWER', 'PESQ', 'PESQi%', 'WER', 'WERi%', 'SASDR', 'PCM', 'SPLERR', 'NLERR',
                   'SPFILE', 'NFILE']
    ti = mixdb.mixtures[mixid].target_file_index[0]
    ni = mixdb.mixtures[mixid].noise_file_index
    metr1 = [mixdb.mixtures[mixid].snr, pesq_mixture, wer_mx, pesq_speech, pesq_impr_pc, wer_tge, wer_pi,
             sa_sdr, pcm, lerr_tg, lerr_n, basename(mixdb.targets[ti].name), basename(mixdb.noises[ni].name)]
    mtab1 = pd.DataFrame([metr1], columns=mtable1_col, index=[mixid])

    # Stats of per frame estimation metrics
    efs_table2_col = ['Max', 'Min', 'Avg', 'Median']
    efs_table2_row = ['PCM', 'SPLERR', 'NLERR']
    metr2 = [[np.max(pcm_frame), np.min(pcm_frame), np.mean(pcm_frame), np.median(pcm_frame)],
             [np.max(lerr_tg_frame), np.min(lerr_tg_frame), np.mean(lerr_tg_frame), np.median(lerr_tg_frame)],
             [np.max(lerr_n_frame), np.min(lerr_n_frame), np.mean(lerr_n_frame), np.median(lerr_n_frame)]]
    mtab2 = pd.DataFrame(metr2, columns=efs_table2_col, index=efs_table2_row)
    mtab2flat_col = ['MXSNR', 'PCM Max', 'PCM Min', 'PCM Avg', 'PCM Median',
                     'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                     'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
    tmp = np.insert(np.array(metr2), 0, mixdb.mixtures[mixid].snr).reshape(1, 13)
    mtab2_flat = pd.DataFrame(tmp, columns=mtab2flat_col, index=[mixid])

    all_metrics_table_1 = mtab1
    all_metrics_table_2 = mtab2_flat

    metric_name = base_name + '_metric_spenh_targetf.txt'
    with open(metric_name, 'w') as f:
        print('Speech enhancement metrics:', file=f)
        print(mtab1.round(2).to_string(), file=f)
        print('', file=f)
        print(f'Extraction statistics over {mixture_f.shape[0]} frames:', file=f)
        print(mtab2.round(2).to_string(), file=f)
        print('', file=f)
        print(f'Target path: {mixdb.targets[ti].name}', file=f)
        print(f'Noise path: {mixdb.noises[ni].name}', file=f)
        # print(f'PESQ improvement: {pesq_impr:0.2f}, {pesq_impr_pc:0.1f}%', file=f)

    lerr_tgtr_frame = None
    lerr_ntr_frame = None
    all_metrics_table_3 = None
    all_metrics_table_4 = None

    if truth_est_mode:
        # cmape_tgtr, cmape_tgtr_bin, cmape_tgtr_frame = mean_abs_percentage_error(hypothesis=target_truth_est_complex,
        #                                                                          reference=truth_f_complex)

        # metrics of estimates using truth instead of prediction
        lerr_tgtr, lerr_tgtr_bin, lerr_tgtr_frame = log_error(truth_f_complex, target_truth_est_complex)
        lerr_ntr, lerr_ntr_bin, lerr_ntr_frame = log_error(noise_f, noise_truth_est_complex)
        ypred_tr_f = np.concatenate((target_truth_est_complex[:, np.newaxis, :],
                                     noise_truth_est_complex[:, np.newaxis, :]), axis=1)
        pcm_tr, pcm_tr_bin, pcm_tr_frame = calc_pcm(hypothesis=ypred_tr_f, reference=ytrue_f, with_log=True)

        # ytrue = np.concatenate((targetfi[:, np.newaxis], noisefi[:, np.newaxis]), axis=1)
        ypred = np.concatenate((target_truth_est_audio[:, np.newaxis], noise_truth_est_audio[:, np.newaxis]), axis=1)
        # scale should be ones
        sa_sdr_tr, opt_scale_tr = calc_sa_sdr(hypothesis=ypred, reference=ytrue, with_scale=True)

        if int(mixdb.mixtures[mixid].snr) > -99:
            pesq_speechtr = calc_pesq(hypothesis=target_truth_est_audio, reference=target)
            # pesq improvement
            pesq_impr_sptr = pesq_speechtr - pesq_mixture
            # pesq improvement %
            pesq_impr_pctr = pesq_impr_sptr / (pesq_mixture + np.finfo(np.float32).eps) * 100
        else:
            pesq_speechtr = 0
            pesq_impr_pctr = 0

        if wer_method == 'none':
            wer_ttge = float('nan')
            wer_tip = float('nan')
        else:
            asr_ttge = calc_asr(target_truth_est_audio, engine=wer_method)
            tmp = calc_wer(asr_ttge.text, asr_tt.text)  # target estimate wer
            wer_ttge = tmp.wer * 100
            wer_tip = (wer_mx - wer_ttge) / (wer_mx + 0.1) * 100

        mtable3_col = ['MXSNR', 'MXPESQ', 'MXWER', 'PESQ', 'PESQi%', 'WER', 'WERi%', 'SASDR', 'PCM', 'SPLERR', 'NLERR']
        metr3 = [mixdb.mixtures[mixid].snr, pesq_mixture, wer_mx, pesq_speechtr, pesq_impr_pctr, wer_ttge,
                 wer_tip, sa_sdr_tr, pcm_tr, lerr_tgtr, lerr_ntr]
        mtab3 = pd.DataFrame([metr3], columns=mtable3_col, index=[mixid])

        # Stats of per frame estimation metrics
        efs_table4_col = ['Max', 'Min', 'Avg', 'Median']
        efs_table4_row = ['PCM', 'SPLERR', 'NLERR']
        metr4 = [[np.max(pcm_tr_frame), np.min(pcm_tr_frame), np.mean(pcm_tr_frame),
                  np.median(pcm_tr_frame)],
                 [np.max(lerr_tgtr_frame), np.min(lerr_tgtr_frame), np.mean(lerr_tgtr_frame),
                  np.median(lerr_tgtr_frame)],
                 [np.max(lerr_ntr_frame), np.min(lerr_ntr_frame), np.mean(lerr_ntr_frame),
                  np.median(lerr_ntr_frame)]]
        mtab4 = pd.DataFrame(metr4, columns=efs_table4_col, index=efs_table4_row)

        # Append extraction metrics to metrics file:
        with open(metric_name, 'a') as f:
            print('', file=f)
            print('Speech enhancement metrics of extraction method using truth:', file=f)
            print(mtab3.round(2).to_string(), file=f)
            print('', file=f)
            print('Extraction (using Truth) statistics over frames:', file=f)
            print(mtab4.round(2).to_string(), file=f)

        # Append to all mixture table
        # mtab4flat_col = ['MXSNR', 'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
        #                  'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
        mtab4flat_col = ['MXSNR', 'PCM Max', 'PCM Min', 'PCM Avg', 'PCM Median',
                         'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                         'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
        # Insert MXSNR
        tmp = np.insert(np.array(metr4), 0, mixdb.mixtures[mixid].snr).reshape(1, 13)
        mtab4_flat = pd.DataFrame(tmp, columns=mtab4flat_col, index=[mixid])

        all_metrics_table_3 = mtab3
        all_metrics_table_4 = mtab4_flat

    # 7) write wav files
    if enable_wav:
        write_wav(name=base_name + '_mixture.wav', audio=float_to_int16(mixture))
        write_wav(name=base_name + '_target_est.wav', audio=float_to_int16(target_est_wav))
        write_wav(name=base_name + '_noise_est.wav', audio=float_to_int16(noise_est_wav))
        write_wav(name=base_name + '_target.wav', audio=float_to_int16(target))
        # write_wav(name=base_name + '_tf_target.wav', audio=float_to_int16(ttf))
        write_wav(name=base_name + '_noise.wav', audio=float_to_int16(noise))
        # debug code to test for perfect reconstruction of the extraction method
        # note both 75% olsa-hanns and 50% olsa-hann modes checked to have perfect reconstruction
        # target_r = mixdb.inverse_transform(target_f)
        # noise_r = mixdb.inverse_transform(noise_f)
        # _write_wav(name=base_name + '_target_r.wav', audio=float_to_int16(target_r))
        # _write_wav(name=base_name + '_noise_r.wav', audio=float_to_int16(noise_r)) # chk perfect rec
        if truth_est_mode:
            write_wav(name=base_name + '_target_truth_est.wav', audio=float_to_int16(target_truth_est_audio))
            write_wav(name=base_name + '_noise_truth_est.wav', audio=float_to_int16(noise_truth_est_audio))

    # 8) Write out plot file
    if enable_plot:
        from matplotlib.backends.backend_pdf import PdfPages
        plot_fname = base_name + '_metric_spenh_targetf.pdf'

        # Reshape feature to eliminate overlap redundancy for easier to understand spectrogram view
        # Original size (frames, stride, num_bands), decimates in stride dimension only if step is > 1
        # Reshape to get frames*decimated_stride, num_bands
        step = int(mixdb.feature_samples / mixdb.feature_step_samples)
        if feature.ndim != 3:
            raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')

        # for feature cn*00n**
        feat_sgram = unstack_complex(feature)
        feat_sgram = 20 * np.log10(abs(feat_sgram) + np.finfo(np.float32).eps)
        feat_sgram = feat_sgram[:, -step:, :]  # decimate,  Fx1xB
        feat_sgram = np.reshape(feat_sgram, (feat_sgram.shape[0] * feat_sgram.shape[1], feat_sgram.shape[2]))

        with PdfPages(plot_fname) as pdf:
            # page1 we always have a mixture and prediction, target optional if truth provided
            tfunc_name = mixdb.targets[0].truth_settings[0].function  # first target, assumes all have same
            if tfunc_name == 'mapped_snr_f':
                # leave as unmapped snr
                predplot = predict
                tfunc_name = mixdb.targets[0].truth_settings[0].function
            elif tfunc_name == 'target_f':
                predplot = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            else:
                # use dB scale
                predplot = 10 * np.log10(predict + np.finfo(np.float32).eps)
                tfunc_name = tfunc_name + ' (db)'

            mixspec = 20 * np.log10(abs(mixture_f) + np.finfo(np.float32).eps)
            pdf.savefig(plot_mixpred(mixture=mixture,
                                     mixture_f=mixspec,
                                     target=target,
                                     feature=feat_sgram,
                                     predict=predplot,
                                     tp_title=tfunc_name))

            # ----- page 2, plot unmapped predict, opt truth reconstructed and line plots of mean-over-f
            # pdf.savefig(plot_pdb_predtruth(predict=pred_snr_f, tp_title='predict snr_f (db)'))

            # page 3 speech extraction
            tg_spec = 20 * np.log10(abs(target_f) + np.finfo(np.float32).eps)
            tg_est_spec = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            # n_spec = np.reshape(n_spec,(n_spec.shape[0] * n_spec.shape[1], n_spec.shape[2]))
            pdf.savefig(plot_epredtruth(predict=tg_est_spec,
                                        predict_wav=target_est_wav,
                                        truth_f=tg_spec,
                                        truth_wav=target,
                                        metric=lerr_tg_frame,
                                        tp_title='speech estimate'))

            # page 4 noise extraction
            n_spec = 20 * np.log10(abs(noise_f) + np.finfo(np.float32).eps)
            n_est_spec = 20 * np.log10(abs(noise_est_complex) + np.finfo(np.float32).eps)
            pdf.savefig(plot_epredtruth(predict=n_est_spec,
                                        predict_wav=noise_est_wav,
                                        truth_f=n_spec,
                                        truth_wav=noisefi,
                                        metric=lerr_n_frame,
                                        tp_title='noise estimate'))

            if truth_est_mode:
                # page 5 truth-based speech extraction
                tg_trest_spec = 20 * np.log10(abs(target_truth_est_complex) + np.finfo(np.float32).eps)
                pdf.savefig(plot_epredtruth(predict=tg_trest_spec,
                                            predict_wav=target_truth_est_audio,
                                            truth_f=tg_spec,
                                            truth_wav=target,
                                            metric=lerr_tgtr_frame,
                                            tp_title='truth-based speech estimate'))

                # page 6 truth-based noise extraction
                n_trest_spec = 20 * np.log10(abs(noise_truth_est_complex) + np.finfo(np.float32).eps)
                pdf.savefig(plot_epredtruth(predict=n_trest_spec,
                                            predict_wav=noise_truth_est_audio,
                                            truth_f=n_spec,
                                            truth_wav=noisefi,
                                            metric=lerr_ntr_frame,
                                            tp_title='truth-based noise estimate'))

        plt.close('all')

    return all_metrics_table_1, all_metrics_table_2, all_metrics_table_3, all_metrics_table_4


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    predict_location = args['--ploc']
    wer_method = args['--wer-method'].lower()
    truth_est_mode = args['--truth-est-mode']
    enable_plot = args['--plot']
    enable_wav = args['--wav']
    enable_summary = args['--summary']
    input_name = args['INPUT']

    from os.path import join

    from tqdm import tqdm

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import p_tqdm_map

    # Setup logging file
    create_file_handler(join(predict_location, 'calc_metric_spenh_targetf.log'))
    update_console_handler(verbose)
    initial_log_messages('calc_metric_spenh_targetf')

    mixdb = MixtureDatabase(config=input_name, show_progress=True)
    mixids = mixdb.mixids_to_list(mixids)
    logger.info(f'Found {len(mixids)} mixtures with {mixdb.num_classes} classes from {input_name}')

    mtab_snr_summary_tr = None
    mtab_snr_summary_emtr = None

    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.predict_location = predict_location
    MP_GLOBAL.truth_est_mode = truth_est_mode
    MP_GLOBAL.enable_plot = enable_plot
    MP_GLOBAL.enable_wav = enable_wav
    MP_GLOBAL.wer_method = wer_method

    progress = tqdm(total=len(mixids), desc='calc_metric_spenh_targetf')
    all_metrics_tables = p_tqdm_map(_process_mixture, mixids, progress=progress)
    progress.close()

    all_metrics_table_1 = pd.concat([item[0] for item in all_metrics_tables])
    all_metrics_table_2 = pd.concat([item[1] for item in all_metrics_tables])
    if truth_est_mode:
        all_metrics_table_3 = pd.concat([item[2] for item in all_metrics_tables])
        all_metrics_table_4 = pd.concat([item[3] for item in all_metrics_tables])

    if not enable_summary:
        return

    # 9) Done with mixtures, write out summary metrics
    # Calculate SNR summary avg of each non-random snr
    all_mtab1_sorted = all_metrics_table_1.sort_values(by=['MXSNR', 'SPFILE'])
    all_mtab2_sorted = all_metrics_table_2.sort_values(by=['MXSNR'])
    mtab_snr_summary = None
    mtab_snr_summary_em = None
    for snri in range(0, len(mixdb.snrs)):
        tmp = all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary = pd.concat([mtab_snr_summary, tmp])

        tmp = all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary_em = pd.concat([mtab_snr_summary_em, tmp])

    # Calculate avg metrics over all mixtures except -99
    all_mtab1_sorted_nom99 = all_mtab1_sorted[all_mtab1_sorted.MXSNR != -99]
    all_nom99_mean = all_mtab1_sorted_nom99.mean(numeric_only=True)

    if truth_est_mode:
        all_mtab3_sorted = all_metrics_table_3.sort_values(by=['MXSNR'])
        all_mtab4_sorted = all_metrics_table_4.sort_values(by=['MXSNR'])
        mtab_snr_summary_tr = all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        mtab_snr_summary_emtr = all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        for snri in range(1, len(mixdb.snrs)):
            mtab_snr_summary_tr = pd.concat([mtab_snr_summary_tr,
                                             all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                 numeric_only=True).to_frame().T])
            mtab_snr_summary_emtr = pd.concat([mtab_snr_summary_emtr,
                                               all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                   numeric_only=True).to_frame().T])

    num_mix = len(mixids)
    if num_mix > 1:
        with open(join(predict_location, 'metric_spenh_targetf_summary.txt'), 'w') as f:
            print(f'Speech enhancement metrics avg over all {len(all_mtab1_sorted_nom99)} non -99 SNR mixtures:',
                  file=f)
            print(all_nom99_mean.to_frame().T.round(2).to_string(), file=f)
            print(f'\nSpeech enhancement metrics avg over each SNR:', file=f)
            print(mtab_snr_summary.round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats avg over each SNR:', file=f)
            print(mtab_snr_summary_em.round(2).to_string(), file=f)
            print('', file=f)

            print(f'Speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_1.describe().round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_2.describe().round(2).to_string(), file=f)
            print('', file=f)

            if truth_est_mode:
                print(f'Truth-based speech enhancement metrics avg over each SNR:', file=f)
                print(mtab_snr_summary_tr.round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistics stats avg over each SNR:', file=f)
                print(mtab_snr_summary_emtr.round(2).to_string(), file=f)
                print('', file=f)

                print(f'Truth-based speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
                print(all_metrics_table_3.describe().round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistic stats over all {num_mix} mixtures:', file=f)
                print(all_metrics_table_4.describe().round(2).to_string(), file=f)
                print('', file=f)

            print('Speech enhancement metrics all-mixtures list:', file=f)
            print(all_metrics_table_1.round(2).to_string(), file=f)
            print('', file=f)
            print('Extraction statistics all-mixtures list:', file=f)
            print(all_metrics_table_2.round(2).to_string(), file=f)

            # Write summary to .csv file
            csv_name = join(predict_location, 'metric_spenh_targetf_summary.csv')
            header_args = {
                'mode': 'a',
                'encoding': 'utf-8',
                'index': False,
                'header': False,
            }
            table_args = {
                'mode': 'a',
                'encoding': 'utf-8',
            }
            label = f'Speech enhancement metrics stats over {num_mix} mixtures:'
            pd.DataFrame([label]).to_csv(csv_name, **header_args)
            all_metrics_table_1.describe().round(2).to_csv(csv_name, encoding='utf-8')
            pd.DataFrame(['']).to_csv(csv_name, **header_args)

            label = f'Extraction statistics stats over {num_mix} mixtures:'
            pd.DataFrame([label]).to_csv(csv_name, **header_args)
            all_metrics_table_2.describe().round(2).to_csv(csv_name, **table_args)
            pd.DataFrame(['']).to_csv(csv_name, **header_args)

            if truth_est_mode:
                label = 'Speech enhancement metrics of extraction method using truth, stats:'
                pd.DataFrame([label]).to_csv(csv_name, **header_args)
                all_metrics_table_3.describe().round(2).to_csv(csv_name, **table_args)
                pd.DataFrame(['']).to_csv(csv_name, **header_args)

                label = 'Truth extraction statistics stats:'
                pd.DataFrame([label]).to_csv(csv_name, **header_args)
                all_metrics_table_4.describe().round(2).to_csv(csv_name, **table_args)
                pd.DataFrame(['']).to_csv(csv_name, **header_args)

            csv_name = join(predict_location, 'metric_spenh_targetf_list.csv')
            pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, **header_args)
            all_metrics_table_1.round(2).to_csv(csv_name, **table_args)

            csv_name = join(predict_location, 'metric_spenh_targetf_estats_list.csv')
            pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, **header_args)
            all_metrics_table_2.round(2).to_csv(csv_name, **table_args)

            if truth_est_mode:
                csv_name = join(predict_location, 'metric_spenh_targetf_truth_list.csv')
                pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, **header_args)
                all_metrics_table_3.round(2).to_csv(csv_name, **table_args)

                csv_name = join(predict_location, 'metric_spenh_targetf_truth_list.csv')
                pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, **header_args)
                all_metrics_table_4.round(2).to_csv(csv_name, **table_args)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from dataclasses import dataclass
from typing import Any

from sonusai.mixture.types import AudioT


@dataclass(frozen=True)
class Data:
    audio: AudioT
    whisper_model: Any = None
    whisper_model_name: str = None
    device: str = None



import numpy as np


def calc_sa_sdr(hypothesis: np.ndarray,
                reference: np.ndarray,
                with_scale: bool = False,
                with_negate: bool = False) -> (np.ndarray, np.ndarray):
    """Calculate source-aggregated SDR (signal distortion ratio) using all source inputs which are [samples, nsrc].

    These should include a noise to be a complete mixture estimate, i.e.,
        noise_est = mixture - sum-over-nsrc(s_est(:, n, :))
    should be one of the sources in s_true and s_est.

    Calculates -10*log10(sumn(||sn||^2) / sumn(||sn - shn||^2)

    Reference:
        SA-SDR: A Novel Loss Function for Separation of Meeting Style Data
        Thilo von Neumann, Keisuke Kinoshita, Christoph Boeddeker, Marc Delcroix, Reinhold Haeb-Umbach
        https://doi.org/10.48550/arXiv.2110.15581

    :param hypothesis: [samples, nsrc]
    :param reference: [samples, nsrc]
    :param with_scale: enable scaling (scaling is same as in SI-SDR)
    :param with_negate: enable negation (for use as a loss function)
    :return: (sa_sdr, opt_scale)
    """

    if with_scale:
        # calc 1 x nsrc scaling factors
        ref_energy = np.sum(reference ** 2, axis=0, keepdims=True)
        # if ref_energy is zero, just set scaling to 1.0
        with np.errstate(divide='ignore', invalid='ignore'):
            opt_scale = np.sum(reference * hypothesis, axis=0, keepdims=True) / ref_energy
            opt_scale[opt_scale == np.inf] = 1.0
            opt_scale = np.nan_to_num(opt_scale, nan=1.0)
        scaled_ref = opt_scale * reference
    else:
        scaled_ref = reference
        opt_scale = np.ones((1, reference.shape[1]), dtype=float)

    # multisrc sa-sdr, inputs must be [samples, nsrc]
    err = scaled_ref - hypothesis

    # -10*log10(sumk(||sk||^2) / sumk(||sk - shk||^2)
    # sum over samples and sources
    num = np.sum(reference ** 2)
    den = np.sum(err ** 2)
    if num == 0 and den == 0:
        ratio = np.inf
    else:
        ratio = num / (den + np.finfo(np.float32).eps)

    sa_sdr = 10 * np.log10(ratio)

    if with_negate:
        # for use as a loss function
        sa_sdr = -sa_sdr

    return sa_sdr, opt_scale



from typing import List

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import AugmentedTargets
from sonusai.mixture.types import ListAudiosT
from sonusai.mixture.types import NoiseFiles


def generate_mixtures(mixdb: MixtureDatabase,
                      augmented_targets: AugmentedTargets,
                      noise_files: NoiseFiles,
                      noise_augmentations: Augmentations,
                      mixups: List[int]) -> int:
    """Generate mixtures and append to mixture database

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param noise_files: List of noise files
    :param noise_augmentations: List of noise augmentations
    :param mixups: List of mixup values
    :return: Number of noise samples used
    """

    from sonusai import SonusAIError

    if mixdb.noise_mix_mode == 'exhaustive':
        return _exhaustive_noise_mix(mixdb=mixdb,
                                     augmented_targets=augmented_targets,
                                     noise_files=noise_files,
                                     noise_augmentations=noise_augmentations,
                                     mixups=mixups)

    if mixdb.noise_mix_mode == 'non-exhaustive':
        return _non_exhaustive_noise_mix(mixdb=mixdb,
                                         augmented_targets=augmented_targets,
                                         mixups=mixups)

    if mixdb.noise_mix_mode == 'non-combinatorial':
        return _non_combinatorial_noise_mix(mixdb=mixdb,
                                            augmented_targets=augmented_targets,
                                            mixups=mixups)

    raise SonusAIError(f'invalid noise_mix_mode: {mixdb.noise_mix_mode}')


def _exhaustive_noise_mix(mixdb: MixtureDatabase,
                          augmented_targets: AugmentedTargets,
                          noise_files: NoiseFiles,
                          noise_augmentations: Augmentations,
                          mixups: List[int]):
    """Exhaustive noise mix mode
    Use every noise/augmentation with every target/augmentation.

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param noise_files: List of noise files
    :param noise_augmentations: List of noise augmentations
    :param mixups: List of mixup values
    :return: Number of noise samples used
    """
    from random import randint

    import numpy as np

    from sonusai.mixture import MRecord
    from sonusai.mixture import evaluate_random_rule
    from sonusai.mixture import get_augmented_target_indices_for_mixup

    used_noise_samples = 0
    for noise_file_index in range(len(noise_files)):
        for noise_augmentation_index, noise_augmentation in enumerate(noise_augmentations):
            noise_offset = 0
            noise_length = len(mixdb.augmented_noise_audios[noise_file_index][noise_augmentation_index])

            for mixup in mixups:
                augmented_target_indices_for_mixup = get_augmented_target_indices_for_mixup(
                        mixdb=mixdb,
                        augmented_targets=augmented_targets,
                        mixup=mixup)

                for augmented_target_index_for_mixup in augmented_target_indices_for_mixup:
                    (target_file_index,
                     target_augmentation_index,
                     target_length) = _get_target_info(
                            mixdb=mixdb,
                            augmented_target_indices_for_mixup=augmented_target_index_for_mixup,
                            augmented_targets=augmented_targets,
                            raw_target_audios=mixdb.raw_target_audios)

                    for spectral_mask_index in range(len(mixdb.spectral_masks)):
                        for snr in mixdb.snrs:
                            mixdb.mixtures.append(MRecord(
                                    name=len(mixdb.mixtures),
                                    target_file_index=target_file_index,
                                    target_augmentation_index=target_augmentation_index,
                                    noise_file_index=noise_file_index,
                                    noise_offset=noise_offset,
                                    noise_augmentation_index=noise_augmentation_index,
                                    samples=target_length,
                                    snr=snr,
                                    spectral_mask_index=spectral_mask_index,
                                    spectral_mask_seed=randint(0, np.iinfo('i').max),
                                    random_snr=False))

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length

                        for random_snr in mixdb.random_snrs:
                            mixdb.mixtures.append(MRecord(
                                    name=len(mixdb.mixtures),
                                    target_file_index=target_file_index,
                                    target_augmentation_index=target_augmentation_index,
                                    noise_file_index=noise_file_index,
                                    noise_offset=noise_offset,
                                    noise_augmentation_index=noise_augmentation_index,
                                    samples=target_length,
                                    snr=evaluate_random_rule(random_snr),
                                    spectral_mask_index=spectral_mask_index,
                                    spectral_mask_seed=randint(0, np.iinfo('i').max),
                                    random_snr=True))

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length
    return used_noise_samples


def _non_exhaustive_noise_mix(mixdb: MixtureDatabase,
                              augmented_targets: AugmentedTargets,
                              mixups: List[int]):
    """Non-exhaustive noise mix mode
    Cycle through every target/augmentation without necessarily using all
    noise/augmentation combinations (reduced data set).

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param mixups: List of mixup values
    :return: Number of noise samples used
    """
    from random import randint

    import numpy as np

    from sonusai.mixture import MRecord
    from sonusai.mixture import evaluate_random_rule
    from sonusai.mixture import get_augmented_target_indices_for_mixup

    used_noise_samples = 0
    noise_offset = 0
    noise_file_index = 0
    noise_augmentation_index = 0
    for mixup in mixups:
        augmented_target_indices_for_mixup = get_augmented_target_indices_for_mixup(
                mixdb=mixdb,
                augmented_targets=augmented_targets,
                mixup=mixup)
        for augmented_target_index_for_mixup in augmented_target_indices_for_mixup:
            (target_file_index,
             target_augmentation_index,
             target_length) = _get_target_info(
                    mixdb=mixdb,
                    augmented_target_indices_for_mixup=augmented_target_index_for_mixup,
                    augmented_targets=augmented_targets,
                    raw_target_audios=mixdb.raw_target_audios)

            for spectral_mask_index in range(len(mixdb.spectral_masks)):
                for snr in mixdb.snrs:
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_offset) = _get_next_noise_offset(target_length=target_length,
                                                            augmented_noise_audios=mixdb.augmented_noise_audios,
                                                            noise_file_index=noise_file_index,
                                                            noise_augmentation_index=noise_augmentation_index,
                                                            noise_offset=noise_offset)

                    mixdb.mixtures.append(MRecord(
                            name=len(mixdb.mixtures),
                            target_file_index=target_file_index,
                            target_augmentation_index=target_augmentation_index,
                            noise_file_index=noise_file_index,
                            noise_augmentation_index=noise_augmentation_index,
                            noise_offset=noise_offset,
                            samples=target_length,
                            snr=snr,
                            spectral_mask_index=spectral_mask_index,
                            spectral_mask_seed=randint(0, np.iinfo('i').max),
                            random_snr=False))

                    noise_offset += target_length
                    used_noise_samples += target_length

                for random_snr in mixdb.random_snrs:
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_offset) = _get_next_noise_offset(target_length=target_length,
                                                            augmented_noise_audios=mixdb.augmented_noise_audios,
                                                            noise_file_index=noise_file_index,
                                                            noise_augmentation_index=noise_augmentation_index,
                                                            noise_offset=noise_offset)

                    mixdb.mixtures.append(MRecord(
                            name=len(mixdb.mixtures),
                            target_file_index=target_file_index,
                            target_augmentation_index=target_augmentation_index,
                            noise_file_index=noise_file_index,
                            noise_augmentation_index=noise_augmentation_index,
                            noise_offset=noise_offset,
                            samples=target_length,
                            snr=evaluate_random_rule(random_snr),
                            spectral_mask_index=spectral_mask_index,
                            spectral_mask_seed=randint(0, np.iinfo('i').max),
                            random_snr=True))

                    noise_offset += target_length
                    used_noise_samples += target_length
    return used_noise_samples


def _non_combinatorial_noise_mix(mixdb: MixtureDatabase,
                                 augmented_targets: AugmentedTargets,
                                 mixups: List[int]):
    """Non-combinatorial noise mix mode
    Combine a target/augmentation with a single cut of a noise/augmentation
    non-exhaustively (each target/augmentation does not use each noise/augmentation).
    Cut has random start and loop back to beginning if end of noise/augmentation is reached.

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param mixups: List of mixup values
    :return: Number of noise samples used
    """
    from random import choice
    from random import randint

    import numpy as np

    from sonusai.mixture import MRecord
    from sonusai.mixture import evaluate_random_rule
    from sonusai.mixture import get_augmented_target_indices_for_mixup

    used_noise_samples = 0
    noise_file_index = 0
    noise_augmentation_index = 0
    for mixup in mixups:
        augmented_target_indices_for_mixup = get_augmented_target_indices_for_mixup(
                mixdb=mixdb,
                augmented_targets=augmented_targets,
                mixup=mixup)
        for augmented_target_index_for_mixup in augmented_target_indices_for_mixup:
            (target_file_index,
             target_augmentation_index,
             target_length) = _get_target_info(
                    mixdb=mixdb,
                    augmented_target_indices_for_mixup=augmented_target_index_for_mixup,
                    augmented_targets=augmented_targets,
                    raw_target_audios=mixdb.raw_target_audios)

            for spectral_mask_index in range(len(mixdb.spectral_masks)):
                for snr in mixdb.snrs:
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_length) = _get_next_noise_indices(augmented_noise_audios=mixdb.augmented_noise_audios,
                                                             noise_file_index=noise_file_index,
                                                             noise_augmentation_index=noise_augmentation_index)

                    mixdb.mixtures.append(MRecord(
                            name=len(mixdb.mixtures),
                            target_file_index=target_file_index,
                            target_augmentation_index=target_augmentation_index,
                            noise_file_index=noise_file_index,
                            noise_augmentation_index=noise_augmentation_index,
                            noise_offset=choice(range(noise_length)),
                            samples=target_length,
                            snr=snr,
                            spectral_mask_index=spectral_mask_index,
                            spectral_mask_seed=randint(0, np.iinfo('i').max),
                            random_snr=False))

                    used_noise_samples += target_length

                for random_snr in mixdb.random_snrs:
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_length) = _get_next_noise_indices(augmented_noise_audios=mixdb.augmented_noise_audios,
                                                             noise_file_index=noise_file_index,
                                                             noise_augmentation_index=noise_augmentation_index)

                    mixdb.mixtures.append(MRecord(
                            name=len(mixdb.mixtures),
                            target_file_index=target_file_index,
                            target_augmentation_index=target_augmentation_index,
                            noise_file_index=noise_file_index,
                            noise_augmentation_index=noise_augmentation_index,
                            noise_offset=choice(range(noise_length)),
                            samples=target_length,
                            snr=evaluate_random_rule(random_snr),
                            spectral_mask_index=spectral_mask_index,
                            spectral_mask_seed=randint(0, np.iinfo('i').max),
                            random_snr=True))

                    used_noise_samples += target_length
    return used_noise_samples


def _get_next_noise_indices(augmented_noise_audios: ListAudiosT,
                            noise_file_index: int,
                            noise_augmentation_index: int) -> (int, int, int):
    noise_augmentation_index += 1
    if noise_augmentation_index == len(augmented_noise_audios[noise_file_index]):
        noise_augmentation_index = 0
        noise_file_index += 1
        if noise_file_index == len(augmented_noise_audios):
            noise_file_index = 0

    return (noise_file_index,
            noise_augmentation_index,
            len(augmented_noise_audios[noise_file_index][noise_augmentation_index]))


def _get_next_noise_offset(target_length: int,
                           augmented_noise_audios: ListAudiosT,
                           noise_file_index: int,
                           noise_augmentation_index: int,
                           noise_offset: int) -> (int, int, int):
    from sonusai import SonusAIError
    if noise_offset + target_length >= len(augmented_noise_audios[noise_file_index][noise_augmentation_index]):
        if noise_offset == 0:
            raise SonusAIError('Length of target audio exceeds length of noise audio')

        noise_offset = 0
        noise_augmentation_index += 1
        if noise_augmentation_index == len(augmented_noise_audios[noise_file_index]):
            noise_augmentation_index = 0
            noise_file_index += 1
            if noise_file_index == len(augmented_noise_audios):
                noise_file_index = 0

    return noise_file_index, noise_augmentation_index, noise_offset


def _get_target_info(mixdb: MixtureDatabase,
                     augmented_target_indices_for_mixup: List[int],
                     augmented_targets: AugmentedTargets,
                     raw_target_audios: AudiosT) -> (List[int], List[int], int):
    from sonusai.mixture import estimate_audio_length
    target_file_index = []
    target_augmentation_index = []
    target_length = 0
    for idx in augmented_target_indices_for_mixup:
        tfi = augmented_targets[idx].target_file_index
        tai = augmented_targets[idx].target_augmentation_index

        target_file_index.append(tfi)
        target_augmentation_index.append(tai)

        raw_target_audio = raw_target_audios[tfi]
        target_augmentation = mixdb.target_augmentations[tai]
        target_length = max(estimate_audio_length(audio=raw_target_audio,
                                                  augmentation=target_augmentation,
                                                  length_common_denominator=mixdb.feature_step_samples),
                            target_length)
    return target_file_index, target_augmentation_index, target_length



"""sonusai keras_predict

usage: keras_predict [-hvr] [-i MIXID] (-m MODEL) (-w KMODEL) [-b BATCH] [-t TSTEPS] INPUT ...

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate if input is a mixture database. [default: *].
    -m MODEL, --model MODEL         Python model file.
    -w KMODEL, --weights KMODEL     Keras model weights file.
    -b BATCH, --batch BATCH         Batch size.
    -t TSTEPS, --tsteps TSTEPS      Timesteps.
    -r, --reset                     Reset model between each file.

Run prediction on a trained Keras model defined by a SonusAI Keras Python model file using SonusAI genft or WAV data.

Inputs:
    MODEL       A SonusAI Python model file with build and/or hypermodel functions.
    KMODEL      A Keras model weights file (or model file with weights).
    INPUT       The input data must be one of the following:
                * Single WAV file or glob of WAV files
                  Using the given model, generate feature data and run prediction. A model file must be
                  provided. The MIXID is ignored.

                * directory
                  Using the given SonusAI mixture database directory, generate feature and truth data if not found.
                  Run prediction. The MIXID is required.

Outputs the following to kpredict-<TIMESTAMP> directory:
    <id>.h5
        dataset:    predict
    keras_predict.log

"""
from typing import Any

from sonusai import logger
from sonusai.mixture import Feature
from sonusai.mixture import Predict


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    model_name = args['--model']
    weights_name = args['--weights']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    reset = args['--reset']
    input_name = args['INPUT']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import isfile
    from os.path import join
    from os.path import splitext

    import h5py
    import keras_tuner as kt
    import tensorflow as tf
    from keras import backend as kb

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import KerasFromH5
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import read_audio
    from sonusai.utils import create_ts_name
    from sonusai.utils import get_frames_per_batch
    from sonusai.utils import import_and_check_keras_model
    from sonusai.utils import reshape_outputs

    if batch_size is not None:
        batch_size = int(batch_size)

    if timesteps is not None:
        timesteps = int(timesteps)

    output_dir = create_ts_name('kpredict')
    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'keras_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('keras_predict')

    logger.info(f'tensorflow    {tf.__version__}')
    logger.info(f'keras         {tf.keras.__version__}')
    logger.info('')

    hypermodel = import_and_check_keras_model(model_name=model_name,
                                              weights_name=weights_name,
                                              timesteps=timesteps,
                                              batch_size=batch_size)
    built_model = hypermodel.build_model(kt.HyperParameters())

    frames_per_batch = get_frames_per_batch(hypermodel.batch_size, hypermodel.timesteps)

    kb.clear_session()
    logger.info('')
    built_model.summary(print_fn=logger.info)
    logger.info('')
    logger.info(f'feature       {hypermodel.feature}')
    logger.info(f'num_classes   {hypermodel.num_classes}')
    logger.info(f'batch_size    {hypermodel.batch_size}')
    logger.info(f'timesteps     {hypermodel.timesteps}')
    logger.info(f'flatten       {hypermodel.flatten}')
    logger.info(f'add1ch        {hypermodel.add1ch}')
    logger.info(f'truth_mutex   {hypermodel.truth_mutex}')
    logger.info(f'input_shape   {hypermodel.input_shape}')
    logger.info('')

    logger.info(f'Loading weights from {weights_name}')
    built_model.load_weights(weights_name)

    logger.info('')
    if len(input_name) == 1 and isdir(input_name[0]):
        input_name = input_name[0]
        logger.info(f'Load mixture database from {input_name}')
        mixdb = MixtureDatabase(config=input_name, show_progress=True)

        if mixdb.feature != hypermodel.feature:
            logger.exception(f'Feature in mixture database does not match feature in model')
            raise SystemExit(1)

        mixids = mixdb.mixids_to_list(mixids)
        if reset:
            # reset mode cycles through each file one at a time
            for mixid in mixids:
                feature, _ = mixdb.mixture_ft(mixid)

                feature, predict = _pad_and_predict(hypermodel=hypermodel,
                                                    built_model=built_model,
                                                    feature=feature,
                                                    frames_per_batch=frames_per_batch)

                output_name = join(output_dir, mixdb.mixtures[mixid].name)
                with h5py.File(output_name, 'a') as f:
                    if 'predict' in f:
                        del f['predict']
                    f.create_dataset(name='predict', data=predict)
        else:
            # Run all data at once using a data generator
            feature = KerasFromH5(mixdb=mixdb,
                                  mixids=mixids,
                                  batch_size=hypermodel.batch_size,
                                  timesteps=hypermodel.timesteps,
                                  flatten=hypermodel.flatten,
                                  add1ch=hypermodel.add1ch)

            predict = built_model.predict(feature, batch_size=hypermodel.batch_size, verbose=1)
            predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)

            # Write data to separate files
            for idx, mixid in enumerate(mixids):
                output_name = join(output_dir, mixdb.mixtures[mixid].name)
                with h5py.File(output_name, 'a') as f:
                    if 'predict' in f:
                        del f['predict']
                    f.create_dataset('predict', data=predict[feature.file_indices[idx]])

        logger.info(f'Saved results to {output_dir}')
        return

    if not all(isfile(file) and splitext(file)[1] == '.wav' for file in input_name):
        logger.exception(f'Do not know how to process input from {input_name}')
        raise SystemExit(1)

    logger.info(f'Run prediction on {len(input_name):,} WAV files')
    for file in input_name:
        # Convert WAV to feature data
        audio = read_audio(file)
        feature = get_feature_from_audio(audio=audio, feature=hypermodel.feature)

        feature, predict = _pad_and_predict(hypermodel=hypermodel,
                                            built_model=built_model,
                                            feature=feature,
                                            frames_per_batch=frames_per_batch)

        output_name = join(output_dir, splitext(basename(file))[0] + '.h5')
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=feature)

            if 'predict' in f:
                del f['predict']
            f.create_dataset(name='predict', data=predict)

    logger.info(f'Saved results to {output_dir}')


def _pad_and_predict(hypermodel: Any, built_model: Any, feature: Feature, frames_per_batch: int) -> (Feature, Predict):
    import numpy as np

    from sonusai.utils import reshape_inputs
    from sonusai.utils import reshape_outputs

    frames = feature.shape[0]
    padding = frames_per_batch - frames % frames_per_batch
    feature = np.pad(array=feature, pad_width=((0, padding), (0, 0), (0, 0)))
    feature, _ = reshape_inputs(feature=feature,
                                batch_size=hypermodel.batch_size,
                                timesteps=hypermodel.timesteps,
                                flatten=hypermodel.flatten,
                                add1ch=hypermodel.add1ch)
    predict = built_model.predict(feature, batch_size=hypermodel.batch_size, verbose=1)
    predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)
    predict = predict[:frames, :]
    return feature, predict


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



"""sonusai genmixdb

usage: genmixdb [-hvmfs] LOC

options:
   -h, --help
   -v, --verbose    Be verbose.
   -m, --mix        Save mixture data. [default: False].
   -f, --ft         Save feature/truth_f data. [default: False].
   -s, --segsnr     Save segsnr data. [default: False].

Create mixture database data for training and evaluation. Optionally, also create mixture audio and feature/truth data.

genmixdb creates a database of training and evaluation feature and truth data generation information. It allows
the choice of audio neural-network feature types that are supported by the Aaware real-time front-end and truth
data that is synchronized frame-by-frame with the feature data.

Here are some examples:

#### Adding target data
Suppose you have an audio file which is an example, or target, of what you want to
recognize or detect. Of course, for training a NN you also need truth data for that
file (also called labels). If you don't already have it, genmixdb can create truth using
energy-based sound detection on each frame of the feature data. You can also select
different feature types. Here's an example:

genmixdb target_gfr32ts2

where target_gfr32ts2 contains config.yml with the following inside:
---
feature: gfr32ts2

targets:
  - name: data/target.wav

target_augmentations:
  - normalize: -3.5
...

The mixture database is written to a JSON file (mixdb.json) in the same directory that contains the config.yml file.

#### Target data mix with noise and augmentation

genmixdb mix_gfr32ts2.yml

where mix_gfr32ts2.yml contains:
---
feature: gfr32ts2

targets:
  - name: data/target.wav

target_augmentations:
  - normalize: -3.5
    pitch: [-3, 0, 3]
    tempo: [0.8, 1, 1.2]

noises:
  - name: data/noise.wav

noise_augmentations:
  - normalize: -3.5

snrs:
  - 20
...

In this example a time-domain mixture is created and feature data is calculated as
specified by 'feature: gfr32ts2'. Various feature types are available which vary in
spectral and temporal resolution (4 ms or higher), and other feature algorithm
parameters. The total feature size, dimension, and #frames for mixture is reported
in the log file (the log file name is genmixdb.log).

Truth (labels) can be automatically created per feature output frame based on sound
energy detection. By default, these are appended to the feature data in a single HDF5
output file. By default, truth/label generation is turned on with default algorithm
and threshold levels (see truth section) and a single class, i.e., detecting a single
type of sound. The truth format is a single float per class representing the
probability of activity/presence, and multi-class truth/labels are possible by
specifying the number of classes and either a scalar index or a vector of indices in
which to put the truth result. For example, 'num_class: 3' and  'truth_index: 2' adds
a 1x3 vector to the feature data with truth put in index 2 (others would be 0) for
data/target.wav being an audio clip from sound type of class 2.

The mixture is created with potential data augmentation functions in the following way:
1. apply noise augmentation rule
2. apply target augmentation rule
3. adjust noise gain for specific SNR
4. add augmented noise to augmented target

The mixture length is the target length by default, and the noise signal is repeated
if it is shorter, or trimmed if longer.

#### Target and noise using path lists

Target and noise audio is specified as a list containing text files, audio files, and
file globs. Text files are processed with items on each line where each item can be a
text file, an audio file, or a file glob. Each item will be searched for audio files
which can be WAV, MP3, FLAC, AIFF, or OGG format with any sample rate, bit depth, or
channel count. All audio files will be converted to 16 kHz, 16-bit, single channel
format before processing. For example,

genmixdb dog-bark.yml

where dog-bark.yml contains:
---
targets:
  - name: slib/dog-outside/*.wav
  - name: slib/dog-inside/*.wav

will find all .wav files in the specified directories and process them as targets.

"""
from dataclasses import dataclass

from sonusai import logger
from sonusai.mixture import Location
from sonusai.mixture import MRecord
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import generate_mixtures


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    mixid_width: int = None
    save_mix: bool = None
    save_ft: bool = None
    save_segsnr: bool = None


MP_GLOBAL = MPGlobal()


def genmixdb(location: Location,
             save_mix: bool = False,
             save_ft: bool = False,
             save_segsnr: bool = False,
             logging: bool = True,
             show_progress: bool = False,
             test_mode: bool = False) -> MixtureDatabase:
    from random import seed

    import numpy as np
    import yaml
    from tqdm import tqdm

    import sonusai
    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import Augmentation
    from sonusai.mixture import MixtureDatabaseConfig
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import TruthSettings
    from sonusai.mixture import balance_targets
    from sonusai.mixture import get_augmentation_indices_for_mixup
    from sonusai.mixture import get_augmentations
    from sonusai.mixture import get_augmented_target_indices_for_mixup
    from sonusai.mixture import get_augmented_targets
    from sonusai.mixture import get_class_weights_threshold
    from sonusai.mixture import get_ir_files
    from sonusai.mixture import get_mixups
    from sonusai.mixture import get_noise_files
    from sonusai.mixture import get_spectral_masks
    from sonusai.mixture import get_target_files
    from sonusai.mixture import load_config
    from sonusai.mixture import log_duration_and_sizes
    from sonusai.mixture import update_truth_settings
    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import seconds_to_hms

    config = load_config(location)

    seed(config['seed'])

    if logging:
        logger.debug(f'Seed: {config["seed"]}')
        logger.debug('Configuration:')
        logger.debug(yaml.dump(config))

    if logging:
        logger.info('Collecting targets')
    target_files = get_target_files(config)
    if len(target_files) == 0:
        raise SonusAIError('Canceled due to no targets')

    if logging:
        logger.debug('List of targets:')
        logger.debug(yaml.dump([target_file.name for target_file in target_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting noises')
    noise_files = get_noise_files(config)
    if logging:
        logger.debug('List of noises:')
        logger.debug(yaml.dump([noise_file.name for noise_file in noise_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting impulse responses')
    ir_files = get_ir_files(config)
    if logging:
        logger.debug('List of impulse responses:')
        logger.debug(yaml.dump([ir_file for ir_file in ir_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting target augmentations')
    target_augmentations = get_augmentations(rules=config['target_augmentations'], num_ir=len(ir_files))
    mixups = get_mixups(target_augmentations)
    if logging:
        for mixup in mixups:
            logger.debug(f'Expanded list of target augmentations for mixup of {mixup}:')
            indices = get_augmentation_indices_for_mixup(target_augmentations, mixup)
            for idx in indices:
                logger.debug(f'- {target_augmentations[idx]}')
            logger.debug('')

    if logging:
        logger.info('Collecting noise augmentations')
    noise_augmentations = get_augmentations(rules=config['noise_augmentations'], num_ir=len(ir_files))
    if logging:
        logger.debug('Expanded list of noise augmentations:')
        for augmentation in noise_augmentations:
            logger.debug(f'- {augmentation}')
        logger.debug('')

    if logging:
        logger.debug(f'SNRs: {config["snrs"]}\n')
        logger.debug(f'Random SNRs: {config["random_snrs"]}\n')
        logger.debug(f'Noise mix mode: {config["noise_mix_mode"]}\n')

    spectral_masks = get_spectral_masks(config)
    if logging:
        logger.debug(f'Spectral masks:')
        for spectral_mask in spectral_masks:
            logger.debug(spectral_mask)
        logger.debug('')

    if config['truth_mode'] not in ['normal', 'mutex']:
        raise SonusAIError(f'invalid truth_mode: {config["truth_mode"]}')
    truth_mutex = config['truth_mode'] == 'mutex'

    if truth_mutex and any(mixup > 1 for mixup in mixups):
        raise SonusAIError(f'Mutex truth mode is not compatible with mixup')

    augmented_targets = get_augmented_targets(target_files, target_augmentations)

    mixdb = MixtureDatabase(
            config=MixtureDatabaseConfig(
                    class_balancing=config['class_balancing'],
                    class_balancing_augmentation=dataclass_from_dict(Augmentation,
                                                                     config['class_balancing_augmentation']),
                    class_labels=config['class_labels'],
                    class_weights_threshold=get_class_weights_threshold(config),
                    noise_mix_mode=config['noise_mix_mode'],
                    feature=config['feature'],
                    first_cba_index=len(target_augmentations),
                    ir_files=ir_files,
                    mixtures=[],
                    noise_augmentations=noise_augmentations,
                    noises=noise_files,
                    num_classes=config['num_classes'],
                    random_snrs=config['random_snrs'],
                    seed=config['seed'],
                    snrs=config['snrs'],
                    spectral_masks=spectral_masks,
                    target_augmentations=target_augmentations,
                    targets=target_files,
                    truth_mutex=truth_mutex,
                    truth_reduction_function=config['truth_reduction_function'],
                    truth_settings=dataclass_from_dict(TruthSettings, update_truth_settings(config['truth_settings']))),
            show_progress=show_progress,
            lazy_load=False)
    mixdb.location = location

    raw_target_audio_samples = sum([len(x) for x in mixdb.raw_target_audios])
    if logging:
        logger.info('')
        logger.info(f'Raw target audio: {seconds_to_hms(seconds=raw_target_audio_samples / SAMPLE_RATE)}')

    noise_audio_samples = sum([len(xa) for x in mixdb.augmented_noise_audios for xa in x])
    if logging:
        logger.info(f'Augmented noise audio: {seconds_to_hms(seconds=noise_audio_samples / SAMPLE_RATE)}')

    augmented_targets = balance_targets(mixdb, augmented_targets)

    used_noise_samples = generate_mixtures(mixdb=mixdb,
                                           augmented_targets=augmented_targets,
                                           noise_files=noise_files,
                                           noise_augmentations=noise_augmentations,
                                           mixups=mixups)

    total_mixtures = len(mixdb.mixtures)
    if logging:
        logger.info('')
        logger.info(f'Found {total_mixtures:,} mixtures to process')

    total_duration = float(sum([mrecord.samples for mrecord in mixdb.mixtures])) / sonusai.mixture.SAMPLE_RATE

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb.num_classes,
                               feature_step_samples=mixdb.feature_step_samples,
                               num_bands=mixdb.fg.num_bands,
                               stride=mixdb.fg.stride,
                               desc='Estimated')
        logger.info(
                f'Feature shape:        {mixdb.fg.stride} x {mixdb.fg.num_bands} ({mixdb.fg.stride * mixdb.fg.num_bands} total params)')
        logger.info(f'Feature samples:      {mixdb.feature_samples} samples ({mixdb.feature_ms} ms)')
        logger.info(f'Feature step samples: {mixdb.feature_step_samples} samples ({mixdb.feature_step_ms} ms)')
        logger.info('')

    # Fill in the details
    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.mixid_width = int(np.ceil(np.log10(len(mixdb.mixtures))))
    MP_GLOBAL.save_mix = save_mix
    MP_GLOBAL.save_ft = save_ft
    MP_GLOBAL.save_segsnr = save_segsnr

    progress = tqdm(total=total_mixtures, desc='genmixdb', disable=not show_progress)
    mixdb.mixtures = p_tqdm_map(_process_mixture, range(total_mixtures), progress=progress)
    progress.close()

    total_samples = mixdb.total_samples()
    total_duration = total_samples / sonusai.mixture.SAMPLE_RATE

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb.num_classes,
                               feature_step_samples=mixdb.feature_step_samples,
                               num_bands=mixdb.fg.num_bands,
                               stride=mixdb.fg.stride,
                               desc='Actual')
        noise_samples_percent = (float(used_noise_samples) / float(noise_audio_samples)) * 100
        logger.info('')
        logger.info(f'Used {noise_samples_percent:,.0f}% of noise audio')
        logger.info('')

    if not test_mode:
        mixdb.save()
        logger.info(f'Wrote mixture database to {mixdb.location}')

    return mixdb


def _process_mixture(mixid: int) -> MRecord:
    from sonusai import SonusAIError
    from sonusai.mixture import initialize_mixture

    mrecord = MP_GLOBAL.mixdb.mixtures[mixid]
    if len(mrecord.target_file_index) != len(mrecord.target_augmentation_index):
        raise SonusAIError('target_file_index and target_augmentation_index are not the same length')

    raw_target_audios = [MP_GLOBAL.mixdb.raw_target_audios[mrecord.target_file_index[ti]] for ti in
                         range(len(mrecord.target_file_index))]
    noise_audio = MP_GLOBAL.mixdb.augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]

    mrecord, genmix_data = initialize_mixture(mixdb=MP_GLOBAL.mixdb,
                                              mrecord=mrecord,
                                              raw_target_audios=raw_target_audios,
                                              noise_audio=noise_audio,
                                              mixid_width=MP_GLOBAL.mixid_width)

    if MP_GLOBAL.save_mix or MP_GLOBAL.save_ft:
        write_data = []

        if MP_GLOBAL.save_mix:
            write_data.append(('targets', genmix_data.targets))
            write_data.append(('noise', genmix_data.noise))
            write_data.append(('mixture', genmix_data.mixture))

        if MP_GLOBAL.save_ft:
            truth_t = MP_GLOBAL.mixdb.mixture_truth_t(mixid=mixid,
                                                      targets=genmix_data.targets,
                                                      noise=genmix_data.noise)
            feature, truth_f = MP_GLOBAL.mixdb.mixture_ft(mixid=mixid,
                                                          mixture=genmix_data.mixture,
                                                          truth_t=truth_t)
            write_data.append(('feature', feature))
            write_data.append(('truth_f', truth_f))

            if MP_GLOBAL.save_segsnr:
                segsnr = MP_GLOBAL.mixdb.mixture_segsnr(mixid=mixid,
                                                        targets=genmix_data.targets,
                                                        noise=genmix_data.noise)
                write_data.append(('segsnr', segsnr))

        MP_GLOBAL.mixdb.write_mixture_data(mixid, write_data)
        MP_GLOBAL.mixdb.write_mixture_metadata(mixid)

    return mrecord


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.utils import seconds_to_hms

    verbose = args['--verbose']
    save_mix = args['--mix']
    save_ft = args['--ft']
    save_segsnr = args['--segsnr']
    location = args['LOC']

    start_time = time.monotonic()

    if exists(location) and not isdir(location):
        remove(location)

    makedirs(location, exist_ok=True)

    create_file_handler(join(location, 'genmixdb.log'))
    update_console_handler(verbose)
    initial_log_messages('genmixdb')

    logger.info(f'Creating mixture database for {location}')
    logger.info('')

    genmixdb(location=location,
             save_mix=save_mix,
             save_ft=save_ft,
             save_segsnr=save_segsnr,
             show_progress=True)

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from sonusai.mixture.truth_functions.data import Data
from sonusai.mixture.types import Truth


def sed(data: Data) -> Truth:
    import numpy as np
    from pyaaware import SED

    from sonusai import SonusAIError

    if data.config.config is None:
        raise SonusAIError('Truth function SED missing config')

    parameters = ['thresholds']
    for parameter in parameters:
        if 'thresholds' not in data.config.config:
            raise SonusAIError(f'Truth function SED config missing required parameter: {parameter}')

    thresholds = data.config.config['thresholds']
    if not _strictly_decreasing(thresholds):
        raise SonusAIError(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')

    if len(data.target_audio) % data.frame_size != 0:
        raise SonusAIError(f'Number of samples in audio is not a multiple of {data.frame_size}')

    # SED wants 1-based indices
    s = SED(thresholds=thresholds,
            index=data.config.index,
            frame_size=data.frame_size,
            num_classes=data.config.num_classes,
            mutex=data.config.mutex)

    target_audio = data.target_audio / data.config.target_gain
    for offset in data.offsets:
        indices = slice(offset, offset + data.frame_size)
        new_truth = s.execute(data.target_fft.energy_t(target_audio[indices]))
        data.truth[indices] = np.reshape(new_truth, (1, len(new_truth)))

    return data.truth


def _strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))



# SonusAI data generator classes
from sonusai.data_generator.keras_from_mixdb import KerasFromH5
from sonusai.data_generator.keras_from_mixdb import KerasFromMixtureDatabase
from sonusai.data_generator.torch_from_mixdb import TorchFromMixtureDatabase



from sonusai.mixture.truth_functions.data import Data
from sonusai.mixture.types import Truth


def phoneme(_data: Data) -> Truth:
    # Read in .txt transcript and run a Python function to generate text grid data
    # (indicating which phonemes are active)
    # Then generate truth based on this data and put in the correct classes based on config.index
    from sonusai import SonusAIError

    raise SonusAIError('Truth function phoneme is not supported yet')



from sonusai.utils import ASRResult
from sonusai.utils.asr_functions.data import Data


def deepgram(data: Data) -> ASRResult:
    import tempfile
    from os import getenv
    from os.path import join

    import magic
    from deepgram import Deepgram
    from deepgram._types import BufferSource

    from sonusai import SonusAIError
    from sonusai.utils import ASRResult
    from sonusai.utils import float_to_int16
    from sonusai.utils import write_wav

    key = getenv('DEEPGRAM_API_KEY')
    if key is None:
        raise SonusAIError('DEEPGRAM_API_KEY environment variable does not exist')

    client = Deepgram(key)
    with tempfile.TemporaryDirectory() as tmp:
        file = join(tmp, 'asr.wav')
        write_wav(name=file, audio=float_to_int16(data.audio))

        mimetype = magic.from_file(file, mime=True)
        with open(file, 'rb') as audio:
            source = BufferSource(buffer=audio, mimetype=mimetype)
            retry = 5
            count = 0
            while True:
                try:
                    results = client.transcription.sync_prerecorded(source)['results']
                    return ASRResult(text=results['channels'][0]['alternatives'][0]['transcript'],
                                     confidence=results['channels'][0]['alternatives'][0]['confidence'])
                except Exception as e:
                    count += 1
                    print(f'Warning: Deepgram exception: {e}')
                    if count >= retry:
                        raise e


"""
Deepgram results:
{'metadata': {'channels': 1,
              'created': '2023-01-30T21:49:44.048Z',
              'duration': 2.3795626,
              'model_info': {'c12089d0-0766-4ca0-9511-98fd2e443ebd': {'name': 'general',
                                                                      'tier': 'base',
                                                                      'version': '2022-01-18.1'}},
              'models': ['c12089d0-0766-4ca0-9511-98fd2e443ebd'],
              'request_id': 'e1154979-07f7-46a3-89e6-d5d796676d31',
              'sha256': '3cad2f30a83e351eab3c4dcaa2ec47185e8f2979c90abec0c2332a7eef7c2d40',
              'transaction_key': 'deprecated'},
 'results': {'channels': [{'alternatives': [{'confidence': 0.9794922,
                                             'transcript': 'the birch can canoe slid on the smooth planks',
                                             'words': [{'confidence': 0.9794922,
                                                        'end': 0.29625,
                                                        'start': 0.13825,
                                                        'word': 'the'},
                                                       {'confidence': 0.9902344,
                                                        'end': 0.57275,
                                                        'start': 0.29625,
                                                        'word': 'birch'},
                                                       {'confidence': 0.73535156,
                                                        'end': 0.73074996,
                                                        'start': 0.57275,
                                                        'word': 'can'},
                                                       {'confidence': 0.9550781,
                                                        'end': 1.08625,
                                                        'start': 0.73074996,
                                                        'word': 'canoe'},
                                                       {'confidence': 0.98876953,
                                                        'end': 1.2442499,
                                                        'start': 1.08625,
                                                        'word': 'slid'},
                                                       {'confidence': 0.9921875,
                                                        'end': 1.3627499,
                                                        'start': 1.2442499,
                                                        'word': 'on'},
                                                       {'confidence': 0.9584961,
                                                        'end': 1.5997499,
                                                        'start': 1.3627499,
                                                        'word': 'the'},
                                                       {'confidence': 0.9970703,
                                                        'end': 1.9947499,
                                                        'start': 1.5997499,
                                                        'word': 'smooth'},
                                                       {'confidence': 0.98828125,
                                                        'end': 2.23175,
                                                        'start': 1.9947499,
                                                        'word': 'planks'}]}]}]}}
"""



from dataclasses import dataclass
from typing import Any
from typing import Union

from sonusai.mixture import AudioT
from sonusai.mixture import Location


@dataclass(frozen=True)
class ASRResult:
    text: str
    confidence: float = None


def calc_asr(audio: Union[AudioT, Location],
             engine: str = 'deepgram',
             whisper_model: Any = None,
             whisper_model_name: str = 'base.en',
             device: str = None) -> ASRResult:
    """Run ASR on audio

    :param audio: Numpy array of audio samples or location of an audio file
    :param engine: Type of ASR engine to use
    :param whisper_model: A preloaded Whisper ASR model
    :param whisper_model_name: Name of Whisper ASR model to use if none was provided
    :param device: the device to put the ASR model into
    :return: ASRResult object containing text and confidence
    """
    from copy import copy

    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import read_audio
    from sonusai.utils import asr_functions
    from sonusai.utils.asr_functions.data import Data

    if not isinstance(audio, np.ndarray):
        audio = copy(read_audio(audio))

    data = Data(audio, whisper_model, whisper_model_name, device)

    try:
        return getattr(asr_functions, engine)(data)
    except AttributeError:
        raise SonusAIError(f'Unsupported ASR function: {engine}')



from sonusai.mixture.truth_functions.data import Data
from sonusai.mixture.types import Truth


def _core(data: Data, mapped: bool, snr: bool) -> Truth:
    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import calculate_mapped_snr_f

    snr_db_mean = None
    snr_db_std = None
    if mapped:
        if data.config.config is None:
            raise SonusAIError('Truth function mapped SNR missing config')

        parameters = ['snr_db_mean', 'snr_db_std']
        for parameter in parameters:
            if parameter not in data.config.config:
                raise SonusAIError(f'Truth function mapped_snr_f config missing required parameter: {parameter}')

        snr_db_mean = data.config.config['snr_db_mean']
        if len(snr_db_mean) != data.target_fft.bins:
            raise SonusAIError(f'Truth function mapped_snr_f snr_db_mean does not have {data.target_fft.bins} elements')

        snr_db_std = data.config.config['snr_db_std']
        if len(snr_db_std) != data.target_fft.bins:
            raise SonusAIError(f'Truth function mapped_snr_f snr_db_std does not have {data.target_fft.bins} elements')

    for index in data.zero_based_indices:
        if index + data.target_fft.bins > data.config.num_classes:
            raise SonusAIError('Truth index exceeds the number of classes')

    for offset in data.offsets:
        target_energy = np.float32(data.target_fft.energy_f(data.target_audio[offset:offset + data.frame_size]))
        if snr:
            noise_energy = np.float32(data.noise_fft.energy_f(data.noise_audio[offset:offset + data.frame_size]))
        else:
            noise_energy = np.float32(1)

        indices = slice(offset, offset + data.frame_size)
        for index in data.zero_based_indices:
            old_err = np.seterr(divide='ignore', invalid='ignore')
            tmp = target_energy / noise_energy
            np.seterr(**old_err)

            tmp = np.nan_to_num(tmp, nan=-np.inf, posinf=np.inf, neginf=-np.inf)

            if mapped:
                tmp = calculate_mapped_snr_f(tmp, snr_db_mean, snr_db_std)

            data.truth[indices, index:index + data.target_fft.bins] = tmp

    return data.truth


def energy_f(data: Data) -> Truth:
    return _core(data=data, mapped=False, snr=False)


def snr_f(data: Data) -> Truth:
    return _core(data=data, mapped=False, snr=True)


def mapped_snr_f(data: Data) -> Truth:
    return _core(data=data, mapped=True, snr=True)


def energy_t(data: Data) -> Truth:
    import numpy as np

    for offset in data.offsets:
        target_energy = data.target_fft.energy_t(data.target_audio[offset:offset + data.frame_size])
        data.truth[offset:offset + data.frame_size, data.zero_based_indices] = np.float32(target_energy)

    return data.truth



from sonusai.mixture.truth_functions.data import Data
from sonusai.mixture.types import Truth


def file(data: Data) -> Truth:
    import h5py
    import numpy as np

    from sonusai import SonusAIError

    if data.config.config is None:
        raise SonusAIError('Truth function file missing config')

    parameters = ['file']
    for parameter in parameters:
        if 'file' not in data.config.config:
            raise SonusAIError(f'Truth function file config missing required parameter: {parameter}')

    with h5py.File(data.config.config['file'], 'r') as f:
        if 'truth_t' not in f:
            raise SonusAIError('Truth file does not contain truth_t dataset')
        truth_in = np.array(f['truth_t'])

    if truth_in.ndim != 2:
        raise SonusAIError('Truth file data is not 2 dimensions')

    if truth_in.shape[0] != len(data.target_audio):
        raise SonusAIError('Truth file does not contain the right amount of samples')

    if len(data.zero_based_indices) > 1:
        if len(data.zero_based_indices) != truth_in.shape[1]:
            raise SonusAIError('Truth file does not contain the right amount of classes')

        data.truth[:, data.zero_based_indices] = truth_in
    else:
        index = data.zero_based_indices[0]
        if index + truth_in.shape[1] > data.config.num_classes:
            raise SonusAIError('Truth file contains too many classes')

        data.truth[:, index:index + truth_in.shape[1]] = truth_in

    return data.truth



import multiprocessing as mp
import warnings
from dataclasses import dataclass
from typing import List

import numpy as np

from sonusai.mixture import Feature
from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Truth

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from keras.utils import Sequence


@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None


MP_GLOBAL = MPGlobal()


def _pool_init(mixdb: MixtureDatabase) -> None:
    MP_GLOBAL.mixdb = mixdb


def _pool_func(mixid: int) -> (Feature, Truth):
    mixdb = MP_GLOBAL.mixdb
    return mixdb.mixture_ft(mixid)


class KerasFromMixtureDatabase(Sequence):
    """Generates data for Keras from a SonusAI mixture database
    """

    @dataclass(frozen=True)
    class BatchParams:
        mixids: List[int]
        offset: int
        extra: int
        padding: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: GeneralizedIDs,
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization
        """
        self.mixdb = mixdb
        self.mixids = self.mixdb.mixids_to_list(mixids)
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle
        self.stride = self.mixdb.fg.stride
        self.num_bands = self.mixdb.fg.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

        self.pool = mp.Pool(processes=mp.cpu_count(),
                            initializer=_pool_init,
                            initargs=[mixdb])

    def __len__(self) -> int:
        """Denotes the number of batches per epoch
        """
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data
        """
        from sonusai.utils import reshape_inputs

        batch_params = self.batch_params[batch_index]

        result = self.pool.map(_pool_func, batch_params.mixids)
        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        pad_shape = list(feature.shape)
        pad_shape[0] = batch_params.padding
        feature = np.vstack([feature, np.zeros(pad_shape)])

        pad_shape = list(truth.shape)
        pad_shape[0] = batch_params.padding
        truth = np.vstack([truth, np.zeros(pad_shape)])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth = reshape_inputs(feature=feature,
                                        truth=truth,
                                        batch_size=self.batch_size,
                                        timesteps=self.timesteps,
                                        flatten=self.flatten,
                                        add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs
        """
        import random

        if self.shuffle:
            random.shuffle(self.mixids)
            self._initialize_mixtures()

    def _initialize_mixtures(self) -> None:
        from sonusai.utils import get_frames_per_batch

        frames_per_batch = get_frames_per_batch(self.batch_size, self.timesteps)
        # Always extend the number of batches to use all available data
        # The last batch may need padding
        self.total_batches = int(np.ceil(self.mixdb.total_feature_frames(self.mixids) / frames_per_batch))

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        self.file_indices = []
        total_frames = 0
        for idx, mixid in enumerate(self.mixids):
            current_frames = self.mixdb.mixture_samples(mixid) // self.mixdb.feature_step_samples
            self.file_indices.append(slice(total_frames, total_frames + current_frames))
            total_frames += current_frames
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                mixids = self.mixids[start_mixture_index:idx + 1]
                self.batch_params.append(self.BatchParams(mixids=mixids, offset=offset, extra=extra, padding=0))
                if extra == 0:
                    start_mixture_index = idx + 1
                    offset = 0
                else:
                    start_mixture_index = idx
                    offset = current_frames - extra
                cumulative_frames = extra

        # If needed, add final batch with padding
        needed_frames = self.total_batches * frames_per_batch
        padding = needed_frames - total_frames
        if padding != 0:
            mixids = self.mixids[start_mixture_index:]
            self.batch_params.append(self.BatchParams(mixids=mixids, offset=offset, extra=0, padding=padding))


KerasFromH5 = KerasFromMixtureDatabase



# SonusAI metrics utilities for model training and validation
from sonusai.metrics.calc_class_weights import calc_class_weights_from_mixdb
from sonusai.metrics.calc_class_weights import calc_class_weights_from_truth
from sonusai.metrics.calc_optimal_thresholds import calc_optimal_thresholds
from sonusai.metrics.calc_pcm import calc_pcm
from sonusai.metrics.calc_pesq import calc_pesq
from sonusai.metrics.calc_sa_sdr import calc_sa_sdr
from sonusai.metrics.calc_sample_weights import calc_sample_weights
from sonusai.metrics.calc_wer import calc_wer
from sonusai.metrics.class_summary import class_summary
from sonusai.metrics.confusion_matrix_summary import confusion_matrix_summary
from sonusai.metrics.one_hot import one_hot
from sonusai.metrics.snr_summary import snr_summary



from typing import List

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthFunctionConfig


def truth_function(target_audio: AudioT, noise_audio: AudioT, config: TruthFunctionConfig) -> Truth:
    from sonusai import SonusAIError
    from sonusai.mixture import truth_functions
    from sonusai.mixture.truth_functions.data import Data

    data = Data(target_audio, noise_audio, config)
    if data.config.target_gain == 0:
        return data.truth

    try:
        return getattr(truth_functions, data.config.function)(data)
    except AttributeError:
        raise SonusAIError(f'Unsupported truth function: {data.config.function}')


def get_truth_indices_for_mixid(mixdb: MixtureDatabase, mixid: int) -> List[int]:
    """Get a list of truth indices for a given mixid."""
    from sonusai.mixture.targets import get_truth_indices_for_target

    indices = []
    for target_file_index in mixdb.mixtures[mixid].target_file_index:
        indices.append(*get_truth_indices_for_target(mixdb.targets[target_file_index]))

    return sorted(list(set(indices)))


def truth_reduction(x: Truth, func: str) -> Truth:
    import numpy as np

    from sonusai import SonusAIError

    if func == 'max':
        return np.max(x, axis=0)

    if func == 'mean':
        return np.mean(x, axis=0)

    raise SonusAIError(f'Invalid truth reduction function: {func}')
